/****************************************************************************/
/*                                                                          */
/* TD_TERM.C                                                                */
/*                                                                          */
/* These routines make up the VT100 terminal emulation for the EnCom demo.  */
/* Please note that the main functions are designed to allow for more than  */
/* one emulation -- we just chose one of the most complex for this demo so  */
/* you wouldn't have to re-invent the wheel.  VT100/ANSI emulation should   */
/* cover just about whatever you need to do, as nearly all emulations are   */
/* subsets of VT100 or simply do the same things with different character   */
/* sequences.                                                               */
/*                                                                          */
/****************************************************************************/
#include "t.h"                                  /* the main TUI header file */
#include "encom.h"                              /* the Encom header file    */

#include "td.h"                                 /* our mini-app header file */
#include "td.def"                               /* the TUICP define file    */
#include "td_globx.h"                           /* externs for the app      */

/*********************/
/* ~init_term_window */
/*                   ********************************************************/
/* This function allocates the user pointer of the window passed, and makes */
/* it into a TERM structure.  If successful, the PORT structure is allocated*/
/* off the TERM structure, and initialized.  The terminal settings are set  */
/* to the defaults, and the status line at the bottom of the screen is      */
/* displayed.                                                               */
/****************************************************************************/
int init_term_window( WINDOW *wnp, int port, long baud )
{
  TERM *termp;
  PORT *portp;

  m_hide();
  wn_att(NORMAL, wnp);
  wn_clear(wnp);                                /* clear the Comm window    */
  if ( (wnp->usr_ptr = calloc(1, sizeof(TERM))) == NULL )
  {
    msg_line("Cannot allocate terminal!", wnp);
    return(0);
  }
  termp             = (TERM *) wnp->usr_ptr;
  termp->norm_att   = wnp->att;
  wnp->eol_wrap     = OFF;
  wnp->cr_lf_flag   = OFF;
  wnp->scroll       = ON;
  wnp->bs_clear     = ON;
  termp->echo       = OFF;
  termp->curr_set   = termp->g0_set = ANSI_S;
  termp->g1_set     = ANSI_G;
  wn_init_tabs( 8, wnp );
  set_term_type(VT100, wnp);

  if ( (termp->portp = (PORT *) calloc(1, sizeof(PORT))) == NULL)
  {
    msg_line("Cannot allocate port!", wnp);
    free(wnp->usr_ptr);
    return(0);
  }
  if ( (termp->xferp = (XFER *) calloc(1, sizeof(XFER))) == NULL)
  {
    msg_line("Cannot allocate port!", wnp);
    free(termp->portp);
    free(wnp->usr_ptr);
    return(0);
  }
  portp = termp->portp;
  if( com_port_create(port, baud, 'N', 8, 1, 2048, 2048, portp) < 0 )
  {
    msg_line("Cannot create Comm Port!", wnp);
    free(termp->xferp);
    free(termp->portp);
    free(wnp->usr_ptr);
    return(0);
  }
  m_show();
  status_line(wnp);
  mv_cs(0, 0, wnp), pl_csr(wnp);
  return(1);
}
/*** end of init_term_window ***/

/*********************/
/* ~kill_term_window */
/*                   ********************************************************/
/****************************************************************************/
void kill_term_window( WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;
  PORT *portp = termp->portp;

  msg_line("EnCom - The communications library of choice!", wnp);
  com_port_destroy(portp);                      /* destroy port             */
  free(termp->xferp);                           /* free the xfer            */
  free(termp->portp);                           /* free the port            */
  free(wnp->usr_ptr);                           /* free the terminal        */
}
/*** end of kill_term_window ***/

/**********************/
/* ~clear_term_window */
/*                    *******************************************************/
/****************************************************************************/
void clear_term_window( void )
{
  int i;

  m_hide();
  for (i=0; i<V_rows - 1; i++)
    mv_cs(0, i, Term_wnp), wn_claol(Term_wnp);
  mv_cs(0, 0, Term_wnp);
  pl_csr(Term_wnp);
  m_show();
}
/*** end of clear_term_window ***/

/***************/
/* ~term_parms */
/*             **************************************************************/
/* Pull up the term parameters dialog and allow the user to change them.    */
/****************************************************************************/
void term_parms( TUI *tuip, int dlg_inx )
{
  TERM *termp = (TERM *) Term_wnp->usr_ptr;
  int ret_inx, end_flag = OFF;

  tone(2048, 5);
  clear_selected(tuip, TERM_CRLF, TERM_OK);

  if (Term_wnp->eol_wrap)
    select(tuip, TERM_EOL_WRAP);
  if (Term_wnp->cr_lf_flag)
    select(tuip, TERM_CRLF);
  if (Term_wnp->scroll)
    select(tuip, TERM_SCROLL);
  if (termp->echo)
    select(tuip, TERM_ECHO);

  center_dialog(tuip, dlg_inx);
  draw_dialog(tuip, dlg_inx);
  while (!end_flag)
  {
    ret_inx = do_dialog(tuip, dlg_inx, DO_NORMAL);
    switch( ret_inx )
    {
      /*---------------- ESC pressed or clicked off dialog -----------------*/
      case TERM_CANCEL:
      case NO_SELECTION:
        end_flag = ON;
        break;

      case TERM_OK:
        end_flag = ON;
        Term_wnp->eol_wrap = (is_selected(tuip, TERM_EOL_WRAP) == F_SELECTED);
        Term_wnp->cr_lf_flag   = (is_selected(tuip, TERM_CRLF) == F_SELECTED);
        Term_wnp->scroll   = (is_selected(tuip, TERM_SCROLL) == F_SELECTED);
        termp->echo        = (is_selected(tuip, TERM_ECHO) == F_SELECTED);
        status_line(Term_wnp);
        break;
    }
  }
  erase_dialog(tuip, dlg_inx);
  pl_csr(Term_wnp);
}
/*** end of term_parms ***/

/*------------------ The following are terminal functions ------------------*/
/*************/
/* ~vt100_ch */
/*           ****************************************************************/
/* This routine processes a character for VT100 terminal emulation.  Please */
/* note this is 100% our own, no ANSI.SYS drivers!                          */
/****************************************************************************/
int vt100_ch( int c, WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */
  int i, p0, p1, processed = ON;

  switch( termp->t_state )
  {
    /*----------------------------------------------------------------------*/
    case 0:                                     /* normal resting state     */
      switch( c )
      {
        case KEY_CTRL_O:                        /* select G0 character set  */
          termp->curr_set = termp->g0_set;
          break;
        case KEY_CTRL_N:                        /* select G1 character set  */
          termp->curr_set = termp->g1_set;
          break;

        case KEY_CTRL_G:                        /* sound the bell           */
          tone(1024, 10);
          break;

        case KEY_CTRL_M:                        /* carriage return          */
          wnp->csr_x = 0;
          if( wnp->cr_lf_flag == OFF )
            break;
        case KEY_CTRL_J:                        /* line feed                */
        case KEY_CTRL_K:
        case KEY_CTRL_L:
          wn_csr_dn( 1, wnp );
          break;

        case '\b':                              /* backspace                */
          wn_bksp( 1, wnp );
          break;

        case '\t':                              /* tab right                */
          wn_tab_right(1, wnp);
          break;
        case 0x1b:                              /* ESCAPE sequence          */
          termp->t_state = 1;
          break;

        default:                                /* not processed, default   */
          processed = OFF;
          break;
      }
      break;

    /*----------------------------------------------------------------------*/
    case 1:                                       /* 1st level escape seq   */
      termp->t_state = 0;
      switch(c)
      {
        case ',': Uw_timers[0] = Tics_per_sec / 4;/* pause for 1/4 sec      */
                  while (Uw_timers[0] > 0) {}
                  break;

        case '7': save_state(wnp);     break;     /* save csr position      */
        case '8': restore_state(wnp);  break;     /* restore csr position   */

        case 'H': wn_set_tab(wnp);     break;     /* set tab at current col */

        case 'E': wnp->csr_x = 0;                 /* start of next line     */
        case 'D': wn_csr_dn( 1, wnp ); break;     /* csr down 1 line        */
        case 'M': wn_csr_up( 1, wnp ); break;     /* cursor up 1 line       */

        case '(': termp->t_state = 3;  break;     /* alternate char set G0  */
        case ')': termp->t_state = 4;  break;     /* alternate char set G1  */

        case '[':                             /* begin parameter sequence   */
          termp->t_state = 2;
          clear_t_parms(termp);
          break;
      }
      break;

    /*----------------------------------------------------------------------*/
    case 2:
      termp->t_state = 0;
      p0 = termp->t_p[0];
      p1 = termp->t_p[1];
      switch( c )
      {
        case 's': save_state(wnp);          break;      /* save csr pos     */
        case 'u': restore_state(wnp);       break;      /* restore csr pos  */

        case 'I': wn_tab_right( p0, wnp );  break;      /* tab right n tabs */
        case 'Z': wn_tab_left(  p0, wnp );  break;      /* tab left n tabs  */

        case 'C': wn_csr_right( p0, wnp );  break;      /* csr right n cols */
        case 'D': wn_csr_left(  p0, wnp );  break;      /* csr left n cols  */

        case 'F': wnp->csr_x = 0;                       /* csr to col 1 &   */
        case 'A': wn_csr_up( p0, wnp );     break;      /* csr up n lines   */
        case 'E': wnp->csr_x = 0;                       /* csr to col 1 &   */
        case 'B': wn_csr_dn( p0, wnp );     break;      /* csr down n lines */

        case 'f':                                       /* csr positioning  */
        case 'H': wn_csr_pos( p0, p1, wnp ); break;
        case 'G':                                       /* csr to col n     */
          wn_csr_pos( wnp->csr_y + 1, p0, wnp );
          break;

        case 'r': wn_scroll_reg( p0, p1, wnp ); break;/* scrolling region   */

        case 'm': set_atts(wnp);     break;           /* set attributes     */
        case 'h': term_ctrl(ON, wnp);  break;         /* terminal modes ON  */
        case 'l': term_ctrl(OFF, wnp); break;         /* terminal modes OFF */

        case '@': wn_ins_del_chars( INSERT, 0, p0, wnp ); break;
        case 'L': wn_ins_del_lines( INSERT, 0, p0, wnp ); break;
        case 'P': wn_ins_del_chars( DELETE, 0, p0, wnp ); break;
        case 'M': wn_ins_del_lines( DELETE, 0, p0, wnp ); break;

        case 'X': wn_cln( p0, wnp );   break;         /* erase n chars      */

        case 'K':                                     /* clr in line        */
          if( p0 == 0 )
            wn_cleol(wnp);
          else if( p0 == 1 )
            wn_clbol(wnp);
          else if( p0 == 2 )
            wn_claol(wnp);
          break;
        case 'J':                                     /* clr in display     */
          if( p0 == 0 )
            wn_cleos(wnp);
          else if( p0 == 1 )
            wn_clbos(wnp);
          else if( p0 == 2 )
            wn_claos(wnp);
          break;

        case 'g':                                     /* clear all tabs     */
          if( p0 == 0 )
            wn_reset_tab(wnp);
          if( p0 == 3 )
            wn_clear_tabs(wnp);
          break;

        case '?':                                     /* set alternate parm */
          termp->t_ps[termp->t_pi] = 1;
          termp->t_state = 2;
          break;
        case ';':                                     /* move to next parm  */
          if( termp->t_pi < 15 )
            termp->t_pi++;
          termp->t_state = 2;
          break;
        default:                                      /* collect parameters */
          termp->t_p[termp->t_pi] *= 10;
          termp->t_p[termp->t_pi] += (uchar) c - '0';
          termp->t_state = 2;
          break;
      }
      break;

    /*----------------------------------------------------------------------*/
    case 3:
      termp->t_state = 0;
      i = (termp->curr_set == termp->g0_set);
      switch( c )
      {
        case '0': termp->g0_set = ANSI_G; break;/* G0 set - ANSI GRAPHICS   */
        case 'A': termp->g0_set = ANSI_U; break;/* G0 set - ANSI UK         */
        case 'B': termp->g0_set = ANSI_S; break;/* G0 set - ANSI STANDARD   */
      }
      if( i )
        termp->curr_set = termp->g0_set;
    break;

    /*----------------------------------------------------------------------*/
    case 4:
      termp->t_state = 0;
      i = (termp->curr_set == termp->g1_set);
      switch( c )
      {
        case '0': termp->g1_set = ANSI_G; break;/* G1 set - ANSI GRAPHICS   */
        case 'A': termp->g1_set = ANSI_U; break;/* G1 set - ANSI UK         */
        case 'B': termp->g1_set = ANSI_S; break;/* G1 set - ANSI STANDARD   */
      }
      if( i )
        termp->curr_set = termp->g1_set;
    break;
  }
  if( !processed )
  {
    if (c & 0x80)
      wn_ch((uchar) c, wnp);
    else
      wn_ch(Char_sets[termp->curr_set][c & 0x7f], wnp);
  }
  return(processed);
}
/*** end of vt100_ch ***/

/**************/
/* ~vt100_key */
/*            ***************************************************************/
/****************************************************************************/
void vt100_key( int c, WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */
  PORT *cp = termp->portp;                      /* get attatched port       */
  char *p = NULL;

  switch( c )
  {
    case    KEY_HOME: p = "\033Oq"; break;                /* upper left     */
    case    KEY_PGUP: p = "\033Os"; break;                /* upper right    */
    case     KEY_END: p = "\033Op"; break;                /* lower left     */
    case    KEY_PGDN: p = "\033On"; break;                /* lower right    */

    case  KEY_INSERT: p = "\033OM"; break;                /* key enter     ?*/
    case  KEY_DELETE: p = "\177";   break;                /* DELETE char    */

    case      KEY_DN: p = termp->csr_key_mode ? "\033OB" : "\033[B"; break;
    case      KEY_UP: p = termp->csr_key_mode ? "\033OA" : "\033[A"; break;
    case    KEY_LEFT: p = termp->csr_key_mode ? "\033OD" : "\033[D"; break;
    case   KEY_RIGHT: p = termp->csr_key_mode ? "\033OC" : "\033[C"; break;

    case      KEY_F1: p = "\033Oy"; break;
    case      KEY_F2: p = "\033OP"; break;
    case      KEY_F3: p = "\033OQ"; break;
    case      KEY_F4: p = "\033OR"; break;
    case      KEY_F5: p = "\033OS"; break;
    case      KEY_F6: p = "\033Ot"; break;
    case      KEY_F7: p = "\033Ou"; break;
    case      KEY_F8: p = "\033Ov"; break;
    case      KEY_F9: p = "\033Ol"; break;
    case     KEY_F10: p = "\033Ow"; break;
    case KEY_CTRL_F1: p = "\033Ox"; break;                /* KEY_F11        */
    default:
      com_putc(c, cp);
      if( termp->echo )
      {
        m_hide();
        vt100_ch( c, wnp );
        if ( !Csr_in_use )
          pl_csr(Term_wnp);
        m_show();
      }
      break;
  }
  if( p != NULL )
  {
    com_puts( (uchar *) p, cp );
    if( termp->echo )
    {
      m_hide();
      while( *p )
        vt100_ch( *p++, wnp );
      if ( !Csr_in_use )
        pl_csr(Term_wnp);
      m_show();
    }
  }
}
/*** end of vt100_key ***/

/*************/
/* ~set_atts */
/*           ****************************************************************/
/****************************************************************************/
uchar T_atts[8] = { BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE };
void wn_fgnd( int att, WINDOW *wnp )
{
  wnp->att &= 0xf8, wnp->att |= (att & 0x07);
}
void wn_bkgnd( int att, WINDOW *wnp )
{
  wnp->att &= 0x8f, wnp->att |= ((att & 0x07) << 4);
}
void set_atts( WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */
  int i = 0;

  do{
    switch( termp->t_p[i] )
    {
      case  0: wnp->att =   termp->norm_att;break;
      case  1: wnp->att |=  HIGH_INT;       break;
      case  2: wnp->att &= ~HIGH_INT;       break;  /* dim                  */
      case  4: wnp->att =   UNDERLINE;      break;
      case  5: wnp->att |=  BLINK;          break;
      case  7: wnp->att =   REVERSE;        break;
      case  8: wnp->att =   INVISIBLE;      break;
      case 22: wnp->att =   termp->norm_att;break;
      case 24: wnp->att =   termp->norm_att;break;   /* underline off       */
      case 25: wnp->att &= ~BLINK;          break;   /* blink     off       */
      case 27: wnp->att =   termp->norm_att;break;   /* reverse   off       */
      case 30: case 31: case 32: case 33:
      case 34: case 35: case 36: case 37:
        wn_fgnd( T_atts[termp->t_p[i] - 30], wnp );
        break;
      case 40: case 41: case 42: case 43:
      case 44: case 45: case 46: case 47:
        wn_bkgnd( T_atts[termp->t_p[i] - 40], wnp );
        break;
    }
  }while( ++i <= termp->t_pi );
}
/*** end of set_atts ***/

/******************/
/* ~clear_t_parms */
/*                ***********************************************************/
/****************************************************************************/
void clear_t_parms( TERM *termp )
{
  int i;

  for( i = 0; i < 16; i++ )
    termp->t_p[i] = termp->t_ps[i] = 0;
  termp->t_pi = 0;
}
/*** clear_t_parms ***/

/**************/
/* ~term_ctrl */
/*            ***************************************************************/
/****************************************************************************/
void term_ctrl( int mode, WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */
  int i = 0;

  do{
    switch( termp->t_p[i] + (termp->t_ps[i] * 1000) )
    {
      case 1000 +  1: termp->csr_key_mode = mode; break;
      case 1000 + 25: csr_style(mode ? C_SBLOCK : C_LINE); break;
      case 1000 +  7: wnp->eol_wrap = mode; break;
    }
  }while( ++i < termp->t_pi );
}
/*** end of term_ctrl ***/

/***************/
/* ~save_state */
/*             **************************************************************/
/****************************************************************************/
void save_state( WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */

  termp->csr_x_save = wnp->csr_x;
  termp->csr_y_save = wnp->csr_y;
  termp->att_save = wnp->att;
  termp->char_set_save = termp->curr_set;
}
/*** end of save_state ***/

/******************/
/* ~restore_state */
/*                ***********************************************************/
/****************************************************************************/
void restore_state( WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */

  wnp->csr_x = termp->csr_x_save;
  wnp->csr_y = termp->csr_y_save;
  wnp->att = termp->att_save;
  termp->curr_set = termp->char_set_save;
}
/*** end of restore_state ***/

/******************/
/* ~set_term_type */
/*                ***********************************************************/
/****************************************************************************/
void set_term_type( int type, WINDOW *wnp )
{
  TERM *termp = (TERM *) wnp->usr_ptr;          /* get attatched terminal   */

  termp->term_type = type;
  wn_scroll_reg( 1, wn_rqty(wnp) - 1, wnp );
  wn_clear(wnp);
  switch( type )
  {
    default:
      termp->vid_func = vt100_ch;
      termp->key_func = vt100_key;
      break;
  }
}
/*** end of set_term_type ***/

/***************/
/* ~term_event */
/*             **************************************************************/
/* This routine processes a terminal event!                                 */
/****************************************************************************/
void term_event( WINDOW *wnp )
{
  TERM  *termp = (TERM *) wnp->usr_ptr;         /* get attatched terminal   */

  switch(Event.key)
  {
    default:
      (*(termp->key_func))(Event.key, wnp);
      break;
  }
}
/*** end of term_event ***/

/*************/
/* ~term_upd */
/*           ****************************************************************/
/* This routine gets characters from the serial port and calls the          */
/* appropriate video function, being sure to hide the mouse before calling  */
/* the output function, and showing it after done!                          */
/****************************************************************************/
int term_upd( void )
{
  TERM  *termp = (TERM *) Term_wnp->usr_ptr;    /* get attatched terminal   */
  PORT  *portp = termp->portp;                  /* get attatched port       */
  XFER  *xferp = termp->xferp;                  /* get attached xfer vars   */
  int   i, c, upd_csr = OFF;

  if (xferp->rx_mode == ASCII)                  /* doing an ASCII download? */
    ascii_download(termp, NULL);
  else if (!xferp->status)
  {
    for( i = 0; (i < 80) && portp->rx_cnt; i++ )
    {
      if (!upd_csr)
        m_hide(), upd_csr = ON;
      c = com_getc(portp);
      (*(termp->vid_func))( c, Term_wnp );
      upd_csr = ON;
    }
    if (upd_csr)
    {
      if ( !Csr_in_use )
        pl_csr(Term_wnp);
      m_show();
    }
  }
  return(1);
}
/*** end of term_upd ***/

/*** END OF FILE ***/